
/// Tag builder functions.
library tags.builder;

import 'dart:html';

/// A function to build an html element.
///
/// An optional build function may be provided to build child elements.
typedef E Tag<E extends Element>([BuildChildren build, args]);

/// A function to build children of a [Tag].
///
/// Each tag function that is called from within a build function will append
/// its element as a child of the current [buildHostNode].  The order of such calls is 
/// equal to the order in which the children are appended.
typedef void BuildChildren();

/// A function to construct an instance of a node.
///
/// All tag functions that are called from within a create function will _not_
/// append any elements to the node being constructed.  The [buildHostNode] will be 
/// _null_ within the scope of this function.  This allows the side-effect free
/// usage of [Tag]s in class initializers.
typedef Node CreateNode();

// Stack of nodes used to build trees.  
// The top of the stack is always the host for the current build scope.
final _scopes = new List<Node>();

// Whether or not the builder is currently in a [Build] function scope.
bool _isBuilding = false;

/// Create a node and build its children.
///
/// The node constructed by [createNode] will be the [buildHostNode] when the 
/// [buildChildren] function is called.  All [Tag] functions that are called 
/// within the scope of [buildChildren] will append their element as a child of 
/// the [buildHostNode].  [Tag] functions called within the scope of 
/// [createNode] are side-effect free.
Node buildNode(CreateNode createNode, BuildChildren buildChildren) {
  final bool outerIsBuilding = _isBuilding;
  var node;
  try {
    _isBuilding = false;
    node = createNode();
    assert(node != null);
  } finally {  
    _isBuilding = outerIsBuilding;  
  }
  if (outerIsBuilding && !_scopes.isEmpty) {
    _scopes.last.append(node);  
  }
  if (buildChildren == null) return node;
  _scopes.add(node);  
  try {
    _isBuilding = true;
    buildChildren();
  } finally {
    _isBuilding = outerIsBuilding;
    _scopes.removeLast();
  }
  return node;
}

/// The node currently being built; may be `null` if none.
Node get buildHostNode {
  if (!_isBuilding || _scopes.isEmpty) return null;
  else return _scopes.last;
}
